@@ -2,6 +2,7 @@
 // Use of this source code is governed by a BSD-style
 // license that can be found in the LICENSE file.
 
+//go:build cgo
 // +build cgo
 
 package main
@@ -13,11 +14,28 @@ import (
 
 func canInternalLink() bool {
 	switch runtime.GOOS {
+	case "aix":
+		return false
 	case "dragonfly":
 		return false
+	case "freebsd":
+		switch runtime.GOARCH {
+		case "arm64":
+			return false
+		}
 	case "linux":
 		switch runtime.GOARCH {
-		case "arm64", "mips64", "mips64le", "mips", "mipsle", "ppc64", "ppc64le":
+		case "arm64", "mips64", "mips64le", "mips", "mipsle", "ppc64", "ppc64le", "riscv64":
+			return false
+		}
+	case "openbsd":
+		switch runtime.GOARCH {
+		case "arm64", "mips64":
+			return false
+		}
+	case "windows":
+		switch runtime.GOARCH {
+		case "arm64":
 			return false
 		}
 	}
